"""
Rules for SNS resources
"""


from cfnlint.rules import CloudFormationLintRule, RuleMatch


class SnsNoRedrivePolicyRule(CloudFormationLintRule):
    """
    Ensure SNS subscriptions have a redrive policy configured
    """

    id = "ES7000"  # noqa: VNE003
    shortdesc = "SNS No Redrive Policy"
    description = "Ensure SNS subscriptions have a redrive policy configured"
    source_url = "https://awslabs.github.io/serverless-rules/rules/sns/redrive_policy/"
    tags = ["sns"]

    _message = "SNS Subscription {} should have a RedrivePolicy property configured."

    def match(self, cfn):
        """
        Match against SNS subscriptions without RedrivePolicy
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::SNS::Subscription"]).items():
            redrive_policy = value.get("Properties", {}).get("RedrivePolicy", None)

            if redrive_policy is None:
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches



"""
Rules for SQS resources
"""


from cfnlint.rules import CloudFormationLintRule, RuleMatch

from ..utils import Value


class SqsNoRedrivePolicyRule(CloudFormationLintRule):
    """
    Ensure SQS queues have a redrive policy configured
    """

    id = "ES6000"  # noqa: VNE003
    shortdesc = "SQS No Redrive Policy"
    description = "Ensure SQS queues have a redrive policy configured"
    source_url = "https://awslabs.github.io/serverless-rules/rules/sqs/redrive_policy/"
    tags = ["sqs"]

    _message = "SQS Queue {} should have a RedrivePolicy property configured."

    def match(self, cfn):
        """
        Match against SQS queues without RedrivePolicy
        """

        matches = {}
        dlqs = []

        for key, value in cfn.get_resources(["AWS::SQS::Queue"]).items():
            redrive_policy = value.get("Properties", {}).get("RedrivePolicy", None)

            if redrive_policy is None:
                matches[key] = RuleMatch(["Resources", key], self._message.format(key))

            else:
                redrive_policy = Value(redrive_policy)
                # If a queue is used as a DLQ, it doesn't need a redrive policy
                # See https://github.com/awslabs/serverless-rules/issues/79
                dlqs.extend(redrive_policy.references)

        return [v for k, v in matches.items() if k not in dlqs]



"""
Rules for Lambda resources
"""


from collections import defaultdict
from typing import Dict, List, Union

from cfnlint.rules import CloudFormationLintRule, RuleMatch

from ..utils import Value


class LambdaTracingRule(CloudFormationLintRule):
    """
    Ensure Lambda functions have tracing enabled
    """

    id = "WS1000"  # noqa: VNE003
    shortdesc = "Lambda Tracing"
    description = "Ensure that Lambda functions have tracing enabled"
    source_url = "https://awslabs.github.io/serverless-rules/rules/lambda/tracing/"
    tags = ["lambda"]

    _message = "Lambda function {} should have TracingConfig.Mode set to 'Active'."

    def match(self, cfn):
        """
        Match against Lambda functions without tracing enabled
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::Lambda::Function"]).items():
            tracing_mode = value.get("Properties", {}).get("TracingConfig", {}).get("Mode", None)

            if tracing_mode != "Active":
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches


class LambdaESMDestinationRule(CloudFormationLintRule):
    """
    Ensure Lambda event source mappings have a destination configured
    """

    id = "ES1001"  # noqa: VNE003
    shortdesc = "Lambda Event Source Mapping Destination"
    description = "Ensure Lambda event source mappings have a destination configured"
    source_url = "https://awslabs.github.io/serverless-rules/rules/lambda/eventsourcemapping_failure_destination/"
    tags = ["lambda"]

    _message = "Lambda event source mapping {} should have a DestinationConfig.OnFailure.Destination property."

    def match(self, cfn):
        """
        Match against Event Source Mappings without a destination configured
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::Lambda::EventSourceMapping"]).items():
            destination = (
                value.get("Properties", {}).get("DestinationConfig", {}).get("OnFailure", {}).get("Destination", False)
            )

            if not destination:
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches


class LambdaPermissionPrincipalsRule(CloudFormationLintRule):
    """
    Ensure that Lambda functions do not have Lambda permissions with different principals
    """

    id = "WS1002"  # noqa: VNE003
    shortdesc = "Lambda Permission Principals"
    description = "Ensure that Lambda functions do not have Lambda permissions with different principals"
    source_url = "https://awslabs.github.io/serverless-rules/rules/lambda/permission_multiple_principals/"
    tags = ["lambda"]
    _message = "Lambda function {} has Lambda permissions with different principals"

    def _get_permissions(self, cfn):
        """
        Parse all AWS::Lambda::Permissions in the template
        """

        permissions = defaultdict(list)

        for _, value in cfn.get_resources(["AWS::Lambda::Permission"]).items():
            principal = Value(value.get("Properties", {}).get("Principal", ""))
            function_name = Value(value.get("Properties", {}).get("FunctionName", ""))

            for reference in function_name.references:
                permissions[reference].append(principal.id)

        return permissions

    def match(self, cfn):
        """
        Match against Lambda functions with multiple Principal for Permissions
        """

        matches = []

        for key, value in self._get_permissions(cfn).items():
            if len(set(value)) > 1:
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches


class LambdaStarPermissionRule(CloudFormationLintRule):
    """
    Ensure that Lambda functions don't have stars in IAM policy actions
    """

    id = "WS1003"  # noqa: VNE003
    shortdesc = "Lambda Star Permission"
    description = "Ensure that Lambda functions don't have stars in IAM policy actions"
    source_url = "https://awslabs.github.io/serverless-rules/rules/lambda/star_permissions/"
    tags = ["lambda"]

    _message = "IAM Role {} with Lambda as principal has policy actions with stars"

    def _get_principals(self, properties) -> List[str]:
        """
        Retrieve principals from assume role policy documents
        """

        principals = []

        for statement in properties.get("AssumeRolePolicyDocument", {}).get("Statement", []):
            if "Service" not in statement.get("Principal", {}):
                continue

            services = statement.get("Principal", {}).get("Service")

            if isinstance(services, str):
                principals.append(services)
            elif isinstance(services, list):
                principals.extend(services)
            # Ignored for now if it's not a list of str

        return principals

    def _get_actions(self, properties) -> List[str]:
        """
        Retrieve actions from policy documents
        """

        actions = []

        for policy in properties.get("Policies", []):
            for statement in policy.get("PolicyDocument", {}).get("Statement", []):

                action = statement.get("Action")

                if isinstance(action, str):
                    actions.append(action)
                elif isinstance(action, list):
                    actions.extend(action)
                # Ignored for now if it's not a list of str

        return actions

    def match(self, cfn):
        """
        Match against IAM roles with Lambda as principal and stars in actions
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::IAM::Role"]).items():
            principals = self._get_principals(value.get("Properties", {}))
            actions = self._get_actions(value.get("Properties", {}))

            if "lambda.amazonaws.com" in principals and any([a == "*" or ":*" in a for a in actions]):
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches


class LambdaLogRetentionRule(CloudFormationLintRule):
    """
    Ensure that Lambda functions have a corresponding Log Group with retention
    """

    id = "WS1004"  # noqa: VNE003
    shortdesc = "Lambda Log Retention"
    description = "Ensure that Lambda functions have a corresponding Log Group with retention"
    source_url = "https://awslabs.github.io/serverless-rules/rules/lambda/log_retention/"
    tags = ["lambda"]

    _message = "Lambda function {} does not have a corresponding log group with a Retention property"

    def _get_valid_functions(self, log_groups):
        """
        Return function names with valid LogGroups
        """

        known_refs = []
        known_names = []

        # Scan log groups for resource names
        for resource in log_groups.values():
            log_group_name = Value(resource.get("Properties", {}).get("LogGroupName", None))
            retention = resource.get("Properties", {}).get("RetentionInDays", None)

            # No retention or log group name, break early
            if log_group_name is None or retention is None:
                continue

            # Searching for references in log group name
            if len(log_group_name.references) > 0:
                known_refs.extend(log_group_name.references)
            # Otherwise, check if this is a hardcoded name
            elif log_group_name.id.find("/aws/lambda/") == 0:
                known_names.append(log_group_name.id[12:])

        return {"ref": known_refs, "name": known_names}

    def match(self, cfn):
        """
        Match against Lambda functions without a LogGroup with a Retention property
        """

        matches = []

        functions = cfn.get_resources(["AWS::Lambda::Function"])
        log_groups = cfn.get_resources(["AWS::Logs::LogGroup"])

        known = self._get_valid_functions(log_groups)

        # Scan functions against log groups
        for function_ref, function in functions.items():
            log_group_found = False

            if function_ref in known["ref"]:
                log_group_found = True

            function_name = function.get("Properties", {}).get("FunctionName", None)
            if function_name is not None and function_name in known["name"]:
                log_group_found = True

            if not log_group_found:
                matches.append(RuleMatch(["Resources", function_ref], self._message.format(function_ref)))

        return matches


class LambdaDefaultMemorySizeRule(CloudFormationLintRule):
    """
    Ensure that Lambda functions have an explicit memory value
    """

    id = "ES1005"  # noqa: VNE003
    shortdesc = "Lambda Default Memory Size"
    description = "Ensure that Lambda functions have an explicit memory value"
    source_url = "https://awslabs.github.io/serverless-rules/rules/lambda/default_memory_size/"
    tags = ["lambda"]

    _message = "Lambda function {} does not have a MemorySize property"

    def match(self, cfn):
        """
        Match against Lambda functions without an explicity MemorySize
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::Lambda::Function"]).items():
            memory_size = value.get("Properties", {}).get("MemorySize", None)

            if memory_size is None:
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches


class LambdaDefaultTimeoutRule(CloudFormationLintRule):
    """
    Ensure that Lambda functions have an explicit timeout value
    """

    id = "ES1006"  # noqa: VNE003
    shortdesc = "Lambda Default Timeout"
    description = "Ensure that Lambda functions have an explicit timeout value"
    source_url = "https://awslabs.github.io/serverless-rules/rules/lambda/default_timeout/"
    tags = ["lambda"]

    _message = "Lambda function {} does not have a Timeout property"

    def match(self, cfn):
        """
        Match against Lambda functions without an explicity Timeout
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::Lambda::Function"]).items():
            timeout = value.get("Properties", {}).get("Timeout", None)

            if timeout is None:
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches


class LambdaAsyncNoDestinationRule(CloudFormationLintRule):
    """
    Ensure that Lambda functions invoked asynchronously have a destination configured
    """

    id = "ES1007"  # noqa: VNE003
    shortdesc = "Lambda Async Destination"
    description = "Ensure that Lambda functions invoked asynchronously have a destination configured"
    source_url = "https://awslabs.github.io/serverless-rules/rules/lambda/async_failure_destination/"
    tags = ["lambda"]

    _message = "Lambda permission {} has an asynchronous permission but doesn't have an EventInvokeConfig resource related to it"  # noqa: E501

    _async_principals = [
        "events.amazonaws.com",
        "events.amazonaws.com.cn",
        "iot.amazonaws.com",
        "iot.amazonaws.com.cn",
        "s3.amazonaws.com",
        "s3.amazonaws.com.cn",
        "sns.amazonaws.com",
        "sns.amazonaws.com.cn",
    ]

    def _get_async_functions(self, permissions: Dict[str, dict]) -> Dict[str, Union[dict, str]]:
        """
        Return a list of FunctionName properties for permissions with principals that invoke Lambda
        functions asynchronously
        """

        async_functions = {}

        for key, value in permissions.items():
            function_name = value.get("Properties", {}).get("FunctionName", "")
            principal = value.get("Properties", {}).get("Principal", "")

            # No FunctionName, we cannot evaluate this rule
            if not function_name:
                continue

            if principal in self._async_principals:
                async_functions[key] = function_name

        return async_functions

    def match(self, cfn):
        """
        Match against Lambda functions without an explicity Timeout
        """

        matches = []

        permissions = cfn.get_resources(["AWS::Lambda::Permission"])
        event_invoke_configs = cfn.get_resources(["AWS::Lambda::EventInvokeConfig"])

        async_functions = self._get_async_functions(permissions)

        found = []

        for value in event_invoke_configs.values():
            function_name = value.get("Properties", {}).get("FunctionName", "")
            on_failure_destination = (
                value.get("Properties", {}).get("DestinationConfig", {}).get("OnFailure", {}).get("Destination", None)
            )

            if function_name in async_functions.values() and on_failure_destination is not None:
                found.append(function_name)

        for key, value in async_functions.items():
            if value not in found:
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches



from .api_gateway import (
    ApiGatewayDefaultThrottlingRule,
    ApiGatewayLoggingRule,
    ApiGatewayStructuredLoggingRule,
    ApiGatewayTracingRule,
)
from .appsync import AppSyncTracingRule
from .eventbridge import EventBridgeDLQRule
from .lambda_ import (
    LambdaAsyncNoDestinationRule,
    LambdaDefaultMemorySizeRule,
    LambdaDefaultTimeoutRule,
    LambdaESMDestinationRule,
    LambdaLogRetentionRule,
    LambdaPermissionPrincipalsRule,
    LambdaStarPermissionRule,
    LambdaTracingRule,
)
from .sns import SnsNoRedrivePolicyRule
from .sqs import SqsNoRedrivePolicyRule
from .step_functions import StepFunctionsTracingRule

__all__ = [
    "ApiGatewayLoggingRule",
    "ApiGatewayStructuredLoggingRule",
    "ApiGatewayDefaultThrottlingRule",
    "ApiGatewayTracingRule",
    "AppSyncTracingRule",
    "EventBridgeDLQRule",
    "LambdaESMDestinationRule",
    "LambdaLogRetentionRule",
    "LambdaPermissionPrincipalsRule",
    "LambdaStarPermissionRule",
    "LambdaTracingRule",
    "LambdaDefaultMemorySizeRule",
    "LambdaDefaultTimeoutRule",
    "LambdaAsyncNoDestinationRule",
    "SnsNoRedrivePolicyRule",
    "SqsNoRedrivePolicyRule",
    "StepFunctionsTracingRule",
]



"""
Rules for AppSync resources
"""


from cfnlint.rules import CloudFormationLintRule, RuleMatch


class AppSyncTracingRule(CloudFormationLintRule):
    """
    Ensure AppSync GraphQL APIs have tracing enabled
    """

    id = "WS3000"  # noqa: VNE003
    shortdesc = "AppSync Tracing"
    description = "Ensure AppSync GraphQL APIs have tracing enabled"
    source_url = "https://awslabs.github.io/serverless-rules/rules/appsync/tracing/"
    tags = ["appsync"]

    _message = "AppSync GraphQL API {} should have XrayEnabled set to true."

    def match(self, cfn):
        """
        Match against AppSync GraphQL Apis without tracing enabled
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::AppSync::GraphQLApi"]).items():
            xray_enabled = value.get("Properties", {}).get("XrayEnabled", False)

            if not xray_enabled:
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches



"""
Utilities 
"""


import re
from typing import List, Tuple, TypeVar, Union

SUB_PATTERN = re.compile(r"\${(?P<ref>[^}]+)}")


TValue = TypeVar("TValue", bound="Value")


class Value:

    id = ""  # noqa: VNE003
    references = None

    def __new__(cls, value: Union[None, dict, str]) -> Union[None, TValue]:
        """
        Create a new Value object

        If the 'value' passed is None, this will return None instead of a class object
        """

        if value is None:
            return None

        return super(Value, cls).__new__(cls)

    def __init__(self, value: Union[dict, str]):
        """
        Parse a CloudFormation value

        This handles intrinsic functions, such as 'Fn::Sub' and 'Fn::Join' and
        returns an object that contains both a uniquely identifiable string and
        references to other resources.
        """

        self.references = []

        # String
        if isinstance(value, str):
            self.id = value

        # Not a dict - return an error here
        elif not isinstance(value, dict):
            raise ValueError("'value' should be of type str or dict, got '%s'" % type(value))

        # 'Ref' intrinsic function
        elif "Ref" in value:
            self.id, self.references = self._get_from_ref(value["Ref"])

        # 'Fn::GetAtt' intrinsic function
        elif "Fn::GetAtt" in value:
            self.id, self.references = self._get_from_getatt(value["Fn::GetAtt"])

        # 'Fn::Join' intrisic function
        elif "Fn::Join" in value:
            self.id, self.references = self._get_from_join(value["Fn::Join"])

        # 'Fn::Sub' intrisic function
        elif "Fn::Sub" in value:
            self.id, self.references = self._get_from_sub(value["Fn::Sub"])

    def _get_from_ref(self, value: str) -> Tuple[str, List[str]]:
        """
        Return the name and references from a 'Ref' intrinsic function
        """

        return [value, [value]]

    def _get_from_getatt(self, value: list) -> Tuple[str, List[str]]:
        """
        Return the name and references from a 'Fn::GetAtt' intrinsic function
        """

        id_ = ".".join(value)
        references = [value[0]]

        return (id_, references)

    def _get_from_join(self, value: list) -> Tuple[str, List[str]]:
        """
        Return the name and references from a 'Fn::Join' intrinsic function
        """

        delimiter = value[0]
        # Using Value() here to get nested references
        sub_values = [Value(v) for v in value[1]]

        id_ = delimiter.join([v.id for v in sub_values])
        references = []
        for sub_value in sub_values:
            references.extend(sub_value.references)

        return (id_, references)

    def _get_from_sub(self, value: Union[str, list]) -> Tuple[str, List[str]]:
        """
        Return the name and references from a 'Fn::Sub' intrinsic function
        """

        pattern = value
        variables = {}

        if isinstance(value, list):
            pattern = value[0]
            # Using Value() here to get nested references
            variables = {k: Value(v) for k, v in value[1].items()}

        references = []

        for match in SUB_PATTERN.findall(pattern):
            if match in variables:
                # Variable with reference(s)
                if len(variables[match].references) > 0:
                    references.extend(variables[match].references)
                # Hard-coded variable
                else:
                    # Replace with hard-coded value in value ID
                    pattern = pattern.replace(f"${{{match}}}", variables[match].id)
            # No matching variable
            else:
                references.append(match)

        return (pattern, references)



"""
Rules for EventBridge resources
"""


from cfnlint.rules import CloudFormationLintRule, RuleMatch


class EventBridgeDLQRule(CloudFormationLintRule):
    """
    Ensure Event Bridge rules have a DLQ configured
    """

    id = "ES4000"  # noqa: VNE003
    shortdesc = "EventBridge DLQ"
    description = "Ensure Event Bridge rules have a DLQ configured"
    source_url = "https://awslabs.github.io/serverless-rules/rules/eventbridge/rule_without_dlq/"
    tags = ["stepfunctions"]

    _message = "EventBridge rule {} should have a DeadLetterConfig.Arn property for all its Targets."

    def match(self, cfn):
        """
        Match against EventBridge rules without a DeadLetterConfig for all targets
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::Events::Rule"]).items():
            targets = value.get("Properties", {}).get("Targets", [])

            if not all(["Arn" in t.get("DeadLetterConfig", {}) for t in targets]):
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cfn_lint_serverless', 'cfn_lint_serverless.rules']

package_data = \
{'': ['*']}

install_requires = \
['cfn-lint>=0.49.2,<1']

setup_kwargs = {
    'name': 'cfn-lint-serverless',
    'version': '0.3.1',
    'description': 'Serverless rules for cfn-lint',
    'long_description': 'cfn-lint-serverless\n===================\n\nRuleset for [cfn-lint](https://github.com/aws-cloudformation/cfn-lint) for serverless applications.\n\nInstallation\n------------\n\n```bash\npip install cfn-lint cfn-lint-serverless\n```\n\nUsage\n-----\n\n```bash\ncfn-lint template.yaml -a cfn_lint_serverless.rules\n```',
    'author': 'Amazon Web Services',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4',
}


setup(**setup_kwargs)



"""
Rules for Step Functions resources
"""


from cfnlint.rules import CloudFormationLintRule, RuleMatch


class StepFunctionsTracingRule(CloudFormationLintRule):
    """
    Ensure Step Functions state machines have tracing enabled
    """

    id = "WS5000"  # noqa: VNE003
    shortdesc = "Step Functions Tracing"
    description = "Ensure that Step Functions state machines have tracing enabled"
    source_url = "https://awslabs.github.io/serverless-rules/rules/step_functions/tracing/"
    tags = ["stepfunctions"]

    _message = "Step Functions state machine {} should have TracingConfiguration.Enabled set to true."

    def match(self, cfn):
        """
        Match against Step Functions state machine without tracing enabled
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::StepFunctions::StateMachine"]).items():
            tracing_enabled = value.get("Properties", {}).get("TracingConfiguration", {}).get("Enabled", False)

            if not tracing_enabled:
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches



"""
Rules for API Gateway resources
"""


import json
import re
from typing import List

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ApiGatewayLoggingRule(CloudFormationLintRule):
    """
    Ensure API Gateway REST and HTTP APIs have logging enabled
    """

    id = "ES2000"  # noqa: VNE003
    shortdesc = "API Gateway Logging"
    description = "Ensure that API Gateway REST and HTTP APIs have logging enabled"
    source_url = "https://awslabs.github.io/serverless-rules/rules/api_gateway/logging/"
    tags = ["apigateway"]

    _message_access_log_settings = "API Gateway stage {} is missing the AccessLogSetting property."
    _message_destination_arn = "API Gateway stage {} is missing the AccessLogSetting.DestinationArn property."

    def match(self, cfn) -> List[RuleMatch]:
        """
        Match against API Gateway stages without log settings
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::ApiGateway::Stage", "AWS::ApiGatewayV2::Stage"]).items():
            if value["Type"] == "AWS::ApiGateway::Stage":
                # REST APIs
                log_settings = value.get("Properties", {}).get("AccessLogSetting", None)
            else:
                # HTTP APIs
                log_settings = value.get("Properties", {}).get("AccessLogSettings", None)

            if log_settings is None:
                matches.append(
                    RuleMatch(
                        ["Resources", key],
                        self._message_access_log_settings.format(key),
                    )
                )
            elif log_settings.get("DestinationArn", None) is None:
                matches.append(RuleMatch(["Resources", key], self._message_destination_arn.format(key)))

        return matches


class ApiGatewayStructuredLoggingRule(CloudFormationLintRule):
    """
    Ensure API Gateway REST and HTTP APIs use structured logging
    """

    id = "WS2001"  # noqa: VNE003
    shortdesc = "API Gateway Structured Logging"
    description = "Ensure that API Gateway REST and HTTP APIs are using structured logs"
    source_url = "https://awslabs.github.io/serverless-rules/rules/api_gateway/structured_logging/"
    tags = ["apigateway"]

    _message = "API Gateway stage {} is not using structured logging for the AccessLogSetting.Format property."
    _log_format_pattern = re.compile(r"\$context\.[a-zA-Z\.]+")

    def _check_log_format(self, log_format: str) -> bool:
        """
        Validate if a log Format is valid JSON

        As users can use non-string keys like '$context.integration.latency', we
        need to transform the log format into something that can be decoded into
        JSON first.
        """

        log_format = self._log_format_pattern.sub("0", log_format)

        try:
            json.loads(log_format)
            return True
        except json.decoder.JSONDecodeError:
            return False

    def match(self, cfn):
        """
        Match against API Gateway stages not using structured logging
        """

        matches = []

        for key, value in cfn.get_resources(["AWS::ApiGateway::Stage", "AWS::ApiGatewayV2::Stage"]).items():
            if value["Type"] == "AWS::ApiGateway::Stage":
                # REST APIs
                log_format = value.get("Properties", {}).get("AccessLogSetting", {}).get("Format", None)
            else:
                # HTTPI APIs
                log_format = value.get("Properties", {}).get("AccessLogSettings", {}).get("Format", None)

            # Ignore if it's not set. Another rule should catch it.
            if log_format is None:
                continue

            if not self._check_log_format(log_format):
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches


class ApiGatewayDefaultThrottlingRule(CloudFormationLintRule):
    """
    Ensure API Gateway REST APIs have throttling enabled
    """

    id = "ES2003"  # noqa: VNE003
    shortdesc = "API Gateway Throttling"
    description = "Ensure that API Gateway REST APIs have default throttling limits set."
    source_url = "https://awslabs.github.io/serverless-rules/rules/api_gateway/default_throttling/"
    tags = ["apigateway"]

    _message_method_settings = "API Gateway stage {} does not have a default MethodSettings property with ThrottlingBurstLimit and ThrottlingRateLimit."  # noqa: E501
    _message_default_route_settings = "API Gateway stage {} does not have a default DefaultRouteSettings property with ThrottlingBurstLimit and ThrottlingRateLimit."  # noqa: E501

    def _match_rest_api(self, key: str, value: dict) -> List[RuleMatch]:
        """
        Match for REST API
        """

        method_settings = [
            ms
            for ms in value.get("Properties", {}).get("MethodSettings", [])
            if ms.get("HttpMethod") == "*" and ms.get("ResourcePath") == "/*"
        ] or [{}]
        default_method_setting = method_settings[0]

        if "ThrottlingBurstLimit" not in default_method_setting or "ThrottlingRateLimit" not in default_method_setting:
            return [RuleMatch(["Resources", key], self._message_method_settings.format(key))]

        return []

    def _match_http_api(self, key: str, value: dict) -> List[RuleMatch]:
        """
        Match for HTTP API
        """

        route_settings = value.get("Properties", {}).get("DefaultRouteSettings", {})

        if "ThrottlingBurstLimit" not in route_settings or "ThrottlingRateLimit" not in route_settings:
            return [RuleMatch(["Resources", key], self._message_default_route_settings.format(key))]

        return []

    def match(self, cfn) -> List[RuleMatch]:
        """
        Match against API Gateway stages without default throttling
        """

        matches = []

        # API Gateway REST APIs
        for key, value in cfn.get_resources(["AWS::ApiGateway::Stage"]).items():
            matches.extend(self._match_rest_api(key, value))

        # API Gateway HTTP APIs
        for key, value in cfn.get_resources(["AWS::ApiGatewayV2::Stage"]).items():
            matches.extend(self._match_http_api(key, value))

        return matches


class ApiGatewayTracingRule(CloudFormationLintRule):
    """
    Ensure API Gateway REST APIs have tracing enabled
    """

    id = "WS2002"  # noqa: VNE003
    shortdesc = "API Gateway Tracing"
    description = "Ensure that API Gateway REST APIs have tracing enabled"
    source_url = "https://awslabs.github.io/serverless-rules/rules/api_gateway/tracing/"
    tags = ["apigateway"]

    _message = "API Gateway stage {} does not have the TracingEnabled property set to true."

    def match(self, cfn):
        """
        Match against API Gateway stages without tracing enabled
        """

        matches = []

        # HTTP APIs don't support X-Ray
        # for key, value in cfn.get_resources(["AWS::ApiGateway::Stage", "AWS::ApiGatewayV2::Stage"]).items():
        for key, value in cfn.get_resources(["AWS::ApiGateway::Stage"]).items():
            tracing_enabled = value.get("Properties", {}).get("TracingEnabled", False)

            if not tracing_enabled:
                matches.append(RuleMatch(["Resources", key], self._message.format(key)))

        return matches
